#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""
A set of methods for reading in data of PDB model cif files and creating an array of
internal representation of bound molecules. The basic use can be as easy as this:

    from pdbeccdutils.core import bm_reader

    bound_molecules = bm_reader.read_pdb_updated_cif_file('/path/to/cif/xxx_updated.cif')
    for i in bound_molecules:
        rdkit_mol = i.component.mol
        rdkit_inchikey = i.component.inchikey_from_rdkit
"""

import os
import rdkit
from rdkit.Chem.rdMolDescriptors import CalcMolFormula
from pdbeccdutils.core import ccd_reader
from pdbeccdutils.core.component import Component

from pdbeccdutils.core.models import (
    CCDProperties,
    ConformerType,
    Descriptor,
    Residue,
    BoundMolecule,
)
from pdbeccdutils.helpers import cif_tools, conversions, mol_tools, helper
from gemmi import cif
from networkx import DiGraph, connected_components


def read_pdb_updated_cif_file(path_to_cif: str, sanitize: bool = True):
    """
    Read in single wwPDB Model CIF and create internal
    representation of its bound-molecules with multiple components.

    Args:
        path_to_cif (str): Path to the cif file
        sanitize (bool): [Defaults: True]

    Raises:
        ValueError: if file does not exist

    Returns:
        A list of CCDResult representations of each bound-molecule.
    """
    if not os.path.isfile(path_to_cif):
        raise ValueError(f"File '{path_to_cif}' does not exists")

    biomolecule_result = []
    bms = infer_bound_molecules(path_to_cif, ["HOH"])
    for bm in bms:
        reader_result = infer_multiple_chem_comp(path_to_cif, bm.to_dict(), sanitize)
        if reader_result:
            biomolecule_result.append(reader_result)

    return biomolecule_result


def infer_multiple_chem_comp(path_to_cif: str, bm: dict, sanitize: bool = True):
    """
    Read in single wwPDB Model CIF and single bound-molecule to create internal
    representation of the bound molecule.

    Args:
        path_to_cif (str): Path to the cif file
        bm (dict): Dictionary representation of bound-molecule
        sanitize (bool): [Defaults: True]


    Returns:
        A dictionary of CCDResult representation of bound-molecule.
    """

    if len(bm["residues"]) <= 1:
        return

    else:
        warnings = []
        errors = []
        sanitized = False
        cif_block = cif.read(path_to_cif).sole_block()
        cif_tools.preprocess_cif_category(cif_block, "_atom_site.")
        cif_tools.preprocess_cif_category(cif_block, "_chem_comp_bond.")

        mol = rdkit.Chem.RWMol()
        index_atoms, bm_atoms_dict = _parse_pdb_atom_site(
            mol, cif_block.get_mmcif_category("_atom_site."), bm
        )
        _parse_pdb_conformers_site(mol, bm_atoms_dict, index_atoms)
        _parse_pdb_bonds_site(
            mol,
            cif_block.get_mmcif_category("_chem_comp_bond."),
            bm_atoms_dict,
            errors,
            index_atoms,
            bm,
        )

        _handle_disconnected_hydrogens(mol)
        if sanitize:
            sanitized = mol_tools.sanitize(mol)

        comp = Component(mol.GetMol(), cif_block)
        descriptors = [
            Descriptor(type="InChI", program="rdkit", value=comp.inchi_from_rdkit),
            Descriptor(
                type="InChIKey", program="rdkit", value=comp.inchikey_from_rdkit
            ),
        ]
        properties = CCDProperties(
            id="",
            name="",
            formula=CalcMolFormula(comp.mol),
            modified_date="",
            pdbx_release_status="",
            weight=round(comp.physchem_properties["exactmw"], 3),
        )
        comp = Component(mol.GetMol(), cif_block, properties, descriptors)
        reader_result = ccd_reader.CCDReaderResult(
            warnings=warnings, errors=errors, component=comp, sanitized=sanitized
        )

        return reader_result


def _handle_disconnected_hydrogens(mol):
    """
    Delete hydrogens without neighbours (degree = 0).
    RDKit works but gives warning for these ("WARNING: not removing hydrogen atom without neighbors").
    However we wouldn't want these hydrogens as they affect molecular properties.

    Args:
        mol (rdkit.Chem.rchem.Mol): Rdkit Mol object with the
            compound representation.
    """
    disconnected_hydrogens = [
        atom
        for atom in mol.GetAtoms()
        if atom.GetAtomicNum() == 1 and atom.GetDegree() == 0
    ]
    atoms_to_remove = [atom.GetIdx() for atom in disconnected_hydrogens]
    atoms_to_remove.sort(reverse=True)

    for atom in atoms_to_remove:
        mol.RemoveAtom(atom)


def _parse_pdb_atom_site(mol, atoms, bm):
    """
    Setup atoms in the bound molecule

    Args:
        mol (rdkit.Chem.rchem.Mol): RDkit Mol object with the
            compound representation.
        atoms (dict): dictionary with parsed _chem_comp_atom
            category.
        bm (dict): dictionary representation of bound-molecule
    """
    if not atoms:
        return

    exclude_alternate_atoms = True
    if exclude_alternate_atoms:
        seen = {}

    index_atoms = []
    bm_atoms_dict = {}

    for i in range(len(atoms["id"])):
        if atoms["group_PDB"][i] != "HETATM":  # only consider nonpolys
            continue
        for j in bm["residues"]:
            if (
                atoms["label_comp_id"][i] == j["label_comp_id"]
                and atoms["auth_asym_id"][i] == j["auth_asym_id"]
                and atoms["auth_seq_id"][i] == j["auth_seq_id"]
            ):
                atom_tuple = (
                    atoms["auth_seq_id"][i],
                    atoms["auth_atom_id"][i],
                    atoms["auth_comp_id"][i],
                )
                if exclude_alternate_atoms:
                    if atom_tuple in seen:
                        continue
                    seen[atom_tuple] = True

                for categories in atoms:
                    if categories not in bm_atoms_dict:
                        bm_atoms_dict[categories] = []
                    else:
                        bm_atoms_dict[categories].append(atoms[categories][i])

                element = atoms["type_symbol"][i]
                element = (
                    element if len(element) == 1 else element[0] + element[1].lower()
                )
                isotope = None

                if element == "D":
                    element = "H"
                    isotope = 2
                elif element == "X":
                    element = "*"

                atom = rdkit.Chem.Atom(element)
                atom.SetProp("name", "/".join(atom_tuple))
                index_atoms.append(
                    (
                        f"{atoms['auth_asym_id'][i]}/{atoms['auth_seq_id'][i]}",
                        atoms["auth_comp_id"][i],
                        atoms["auth_atom_id"][i],
                    )
                )

                if isotope is not None:
                    atom.SetIsotope(isotope)

                mol.AddAtom(atom)

    return index_atoms, bm_atoms_dict


def _parse_pdb_conformers_site(mol, atoms, index_atoms):
    """
    Setup model cooordinates in the rdkit Mol object.

    Args:
        mol (rdkit.Chem.rdchem.Mol): RDKit Mol object with the compound
            representation.
        atoms (dict): mmcif category with atom info category.
        index_atoms (list): List of intx atoms
    """
    if not atoms:
        return

    model = _setup_pdb_conformer_site(
        atoms, "Cartn_{}", ConformerType.Model.name, index_atoms
    )

    mol.AddConformer(model, assignId=True)


def _setup_pdb_conformer_site(atoms, label, name, index_atoms):
    """
    Setup a conformer

    Args:
        atoms (dict): mmcif category with the atom info.
        label (str): Namespace with the [x,y,z] coordinates.
        name (str): Conformer name.
        index_atoms (lsit): List of index atoms

    Returns:
        rdkit.Chem.rdchem.Conformer: Conformer of the component.
    """
    if not atoms:
        return

    conformer = rdkit.Chem.Conformer(len(index_atoms))

    j = 0
    for i in range(len(atoms["id"])):
        if atoms["group_PDB"][i] != "HETATM":
            continue
        x = conversions.str_to_float(atoms[label.format(("x"))][i])
        y = conversions.str_to_float(atoms[label.format(("y"))][i])
        z = conversions.str_to_float(atoms[label.format(("z"))][i])
        index = helper.find_element_in_list(
            index_atoms,
            (
                f"{atoms['auth_asym_id'][i]}/{atoms['auth_seq_id'][i]}",
                atoms["auth_comp_id"][i],
                atoms["auth_atom_id"][i],
            ),
        )

        atom_position = rdkit.Chem.rdGeometry.Point3D(x, y, z)
        conformer.SetAtomPosition(index, atom_position)
        j = j + 1

        conformer.SetProp("name", name)

    return conformer


def _parse_pdb_bonds_site(mol, bonds, atoms, errors, index_atoms, bm):
    """
    Setup bonds in the compound

    Args:
        mol (rdkit.Chem.rdchem.Mol): Molecule which receives bonds.
        bonds (dict): mmcif category with the bonds info. TODO not being used
        atoms (dict): mmcif category with the atom info. TODO not being used
        errors (list[str]): Issues encountered while parsing.
        index_atoms (list): List of index atoms
        bm (dict): Dictionary representation of bound-molecule
    """
    if not bonds:
        return

    bond_per_type = {}
    for i in range(len(bonds["atom_id_1"])):
        content = (
            bonds["atom_id_1"][i],
            bonds["atom_id_2"][i],
            bonds["value_order"][i],
        )
        if bonds["comp_id"][i] in bond_per_type:
            bond_per_type[bonds["comp_id"][i]].append(content)
        else:
            bond_per_type[bonds["comp_id"][i]] = [content]

    # get connection mapping
    connection_mapping = {}
    for j in bm["residues"]:
        connection_mapping[j["id"]] = j

    for j in bm["residues"]:
        lig = j["label_comp_id"]
        if lig not in bond_per_type:  # ions
            continue
        chain_res = f"{j['auth_asym_id']}/{j['auth_seq_id']}"
        try:
            for pairs in bond_per_type[lig]:
                tuple_to_find_1 = (chain_res, lig, pairs[0])
                tuple_to_find_2 = (chain_res, lig, pairs[1])
                atom_1 = helper.find_element_in_list(index_atoms, tuple_to_find_1)
                atom_2 = helper.find_element_in_list(index_atoms, tuple_to_find_2)
                bond_order = ccd_reader._bond_pdb_order(pairs[2])
                if any(a is None for a in [atom_1, atom_2, bond_order]):
                    pass
                else:
                    mol.AddBond(atom_1, atom_2, bond_order)
        except ValueError:
            errors.append(
                f"Error perceiving {atom_1} - {atom_2} bond in _chem_comp_bond"
            )
        except RuntimeError:
            errors.append(f"Duplicit bond {atom_1} - {atom_2}")

    # Add bound molecule connections
    try:
        for connection in bm["connections"]:
            first_res = connection_mapping[connection[0]]
            second_res = connection_mapping[connection[1]]
            chain_res_1 = f"{first_res['auth_asym_id']}/{first_res['auth_seq_id']}"
            chain_res_2 = f"{second_res['auth_asym_id']}/{second_res['auth_seq_id']}"
            lig_1 = first_res["label_comp_id"]
            lig_2 = second_res["label_comp_id"]
            tuple_to_find_1 = (chain_res_1, lig_1, connection[2]["atom_id_1"])
            tuple_to_find_2 = (chain_res_2, lig_2, connection[2]["atom_id_2"])
            atom_1 = helper.find_element_in_list(index_atoms, tuple_to_find_1)
            atom_2 = helper.find_element_in_list(index_atoms, tuple_to_find_2)

            if any(a is None for a in [atom_1, atom_2]):
                pass
            else:
                mol.AddBond(atom_1, atom_2, ccd_reader._bond_pdb_order("SING"))
    except ValueError:
        errors.append(f"Error perceiving {atom_1} - {atom_2} bond in _struct_conn")
    except RuntimeError:
        errors.append(f"Duplicit bond {atom_1} - {atom_2}")


def infer_bound_molecules(structure, to_discard):
    """Identify bound molecules in the input protein structure.

    Args:
        structure (str): Path to the structure.
        to_discard (list of str): List of residue names to be discarded
    """

    bms = parse_bound_molecules(structure, to_discard)
    bound_molecules = []

    for bm_nodes in connected_components(bms.to_undirected()):
        subgraph = bms.subgraph(bm_nodes)
        bm = BoundMolecule(subgraph)
        bound_molecules.append(bm)

    bound_molecules = sorted(bound_molecules, key=lambda l: -len(l.graph.nodes))
    return bound_molecules


def __add_connections(struct_conn, bms):
    for i in range(len(struct_conn["id"])):
        (ptnr1, atom1) = find_pntr_entry(struct_conn, bms.nodes, 1, i)
        (ptnr2, atom2) = find_pntr_entry(struct_conn, bms.nodes, 2, i)

        # we want covalent connections among ligands only.
        if ptnr1 and ptnr2 and struct_conn["conn_type_id"][i] != "metalc":
            bms.add_edge(ptnr1, ptnr2, atom_id_1=atom1, atom_id_2=atom2)


def __add_con_branch_link(entity_branch_link, branch_scheme, bms):
    entities = {}
    for i in range(len(branch_scheme["entity_id"])):
        if branch_scheme["entity_id"][i] not in entities:
            entities[branch_scheme["entity_id"][i]] = [branch_scheme["pdb_asym_id"][i]]
        elif (
            branch_scheme["pdb_asym_id"][i]
            not in entities[branch_scheme["entity_id"][i]]
        ):
            entities[branch_scheme["entity_id"][i]].append(
                branch_scheme["pdb_asym_id"][i]
            )

    for i in range(len(entity_branch_link["link_id"])):
        ent_id = entity_branch_link["entity_id"][i]
        for chain in entities[ent_id]:
            for node in bms.nodes:
                prtnr1 = False
                prtnr2 = False
                atom1 = False
                atom2 = False
                if (
                    node.name == entity_branch_link["comp_id_1"][i]
                    and node.chain == chain
                    and node.res_id == entity_branch_link["entity_branch_list_num_1"][i]
                ):
                    prtnr1 = node
                    atom1 = node.name
                elif (
                    node.name == entity_branch_link["comp_id_2"][i]
                    and node.chain == chain
                    and node.res_id == entity_branch_link["entity_branch_list_num_2"][i]
                ):
                    prtnr2 = node
                    atom2 = node.name
            if prtnr1 and prtnr2:
                bms.add_edge(prtnr1, prtnr2, atom_id_1=atom1, atom_id_2=atom2)


def parse_bound_molecules(path, to_discard):
    """Parse information from the information about HETATMS from the
    `_pdbx_nonpoly_scheme` and connectivity among them from `_struct_conn`.

    Args:
        path (str): Path to the mmCIF structure
        to_discard (list of str): List of residue names to be discarded.

    Returns:
        DiGraph: All the bound molecules in a given entry.
    """
    doc = cif.read(path)
    cif_block = doc.sole_block()

    if (
        "_pdbx_nonpoly_scheme." not in cif_block.get_mmcif_category_names()
        and "_pdbx_branch_scheme." not in cif_block.get_mmcif_category_names()
    ):
        return DiGraph()

    if "_pdbx_nonpoly_scheme." in cif_block.get_mmcif_category_names():
        bms = parse_ligands_from_nonpoly_scheme(
            cif_block.get_mmcif_category("_pdbx_nonpoly_scheme."), to_discard
        )

    else:
        bms = DiGraph()

    if "_pdbx_branch_scheme." in cif_block.get_mmcif_category_names():
        if bms:
            bms = parse_ligands_from_branch_scheme(
                cif_block.get_mmcif_category("_pdbx_branch_scheme."), to_discard, bms
            )
        else:
            bms = parse_ligands_from_branch_scheme(
                cif_block.get_mmcif_category("_pdbx_branch_scheme."),
                to_discard,
                DiGraph(),
            )

    if "_struct_conn." in cif_block.get_mmcif_category_names():
        __add_connections(cif_block.get_mmcif_category("_struct_conn."), bms)

    if "_pdbx_branch_scheme." in cif_block.get_mmcif_category_names():
        __add_con_branch_link(
            cif_block.get_mmcif_category("_pdbx_entity_branch_link."),
            cif_block.get_mmcif_category("_pdbx_branch_scheme."),
            bms,
        )

    return bms


def find_pntr_entry(struct_conn, residue_pool, partner, i):
    """Helper method to find ligand residue in parsed ligands and check
    its connections.

    Just because it is chatty and used throughout the code quite often.

    Args:
        struct_conn (dict of str: str): struct_conn table.
        residue_pool (List of Residue): List of all the parsed residues.
        partner (str): Identification of the partner (should be 1 or 2)
        i (int): Index in the struct_conn table
    """
    atom_name = struct_conn[f"ptnr{partner}_label_atom_id"][i]
    for residue in residue_pool:
        if (
            residue.name == struct_conn[f"ptnr{partner}_label_comp_id"][i]
            and residue.chain == struct_conn[f"ptnr{partner}_auth_asym_id"][i]
            and residue.res_id == struct_conn[f"ptnr{partner}_auth_seq_id"][i]
        ):
            return (residue, atom_name)

    return (None, atom_name)


def parse_ligands_from_branch_scheme(branch_scheme, to_discard, g):
    """Parse ligands from the mmcif file.

    Args:
        nonpoly_scheme (dict of str: list of str): mmcif _nonpoly_scheme
            category.
        to_discard (list of str): List of residue names to be discarded.

    Returns:
        DiGraph: Ligands and their connectivity in a PDB entry
    """

    for i in range(len(branch_scheme["asym_id"])):
        n = Residue(
            branch_scheme["pdb_mon_id"][i],  # aka label_comp_id
            branch_scheme["pdb_asym_id"][i],  # aka auth_asym_id
            branch_scheme["num"][i],  # aka auth_seq_id
            None,  # aka pdbx_PDB_ins_code
            branch_scheme["entity_id"][i],
        )

        if n.name not in to_discard:
            g.add_node(n)

    return g


def parse_ligands_from_nonpoly_scheme(nonpoly_scheme, to_discard):
    """Parse ligands from the mmcif file.

    Args:
        nonpoly_scheme (dict of str: list of str): mmcif _nonpoly_scheme
            category.
        to_discard (list of str): List of residue names to be discarded.

    Returns:
        DiGraph: Ligands and their connectivity in a PDB entry
    """
    g = DiGraph()

    for i in range(len(nonpoly_scheme["asym_id"])):
        n = Residue(
            nonpoly_scheme["pdb_mon_id"][i],  # aka label_comp_id
            nonpoly_scheme["pdb_strand_id"][i],  # aka auth_asym_id
            nonpoly_scheme["pdb_seq_num"][i],  # aka auth_seq_id
            nonpoly_scheme["pdb_ins_code"][i],  # aka pdbx_PDB_ins_code
            nonpoly_scheme["entity_id"][i],
        )

        if n.name not in to_discard:
            g.add_node(n)

    return g



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Structure writing module. Presently the following formats are supported:
    SDF, CIF, PDB, JSON, XYZ, XML, CML.

Raises:
    CCDUtilsError: If deemed format is not supported or an unrecoverable
        error occurres.
"""
import json
import logging
import math
from pathlib import Path
from typing import List
from xml.dom import minidom
from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element, SubElement

import pdbeccdutils
import rdkit
import gemmi
from gemmi import cif
from pdbeccdutils.helpers import cif_tools
from pdbeccdutils.core.component import Component
from pdbeccdutils.core.exceptions import CCDUtilsError
from pdbeccdutils.core.models import ConformerType


def write_molecule(
    path,
    component: Component,
    remove_hs: bool = True,
    alt_names: bool = False,
    conf_type: ConformerType = ConformerType.Ideal,
):
    """Export molecule in a specified format. Presently supported formats
    are: PDB CCD CIF (*.cif); Mol file (*.sdf); Chemical Markup language
    (*.cml); PDB file (*.pdb); XYZ file (*.xyz); XML (*.xml).
    ConformerType.AllConformers is presently supported only for PDB.

    Args:
        path (str|Path): Path to the file. Suffix determines format to be
            used.
        component (Component): Component to be exported
        remove_hs (bool, optional): Defaults to True. Whether or not
            hydrogens should be removed.
        alt_names (bool, optional): Defaults to False. Whether or not
            alternate names should be exported.
        conf_type (ConformerType, optional):
            Defaults to ConformerType.Ideal. Conformer type to be
            exported.

    Raises:
        CCDUtilsError: For unsupported format
    """
    path = str(path) if isinstance(path, Path) else path

    extension = path.split(".")[-1].lower()
    str_representation = ""

    if extension in ("sdf", "mol"):
        str_representation = to_sdf_str(component, remove_hs, conf_type)
    elif extension == "pdb":
        str_representation = to_pdb_str(component, remove_hs, alt_names, conf_type)
    elif extension in ("mmcif", "cif"):
        to_pdb_ccd_cif_file(path, component, remove_hs)
        return
    elif extension == "cml":
        str_representation = to_cml_str(component, remove_hs, conf_type)
    elif extension == "xml":
        str_representation = to_xml_str(component, remove_hs, conf_type)
    elif extension == "xyz":
        str_representation = to_xyz_str(component, remove_hs, conf_type)
    elif extension == "json":
        str_representation = json.dumps(
            to_json_dict(component, remove_hs, conf_type), sort_keys=True, indent=4
        )
    else:
        raise CCDUtilsError("Unsupported file format: {}".format(extension))

    with open(path, "w") as f:
        f.write(str_representation)


def to_pdb_str(
    component: Component,
    remove_hs: bool = True,
    alt_names: bool = False,
    conf_type: ConformerType = ConformerType.Ideal,
):
    """Converts structure to the PDB format.

    Args:
        Component (Component): Component to be exported.
        remove_hs (bool, optional): Defaults to True.
        alt_names (bool, optional): Defaults to False. Whether or not
            alternate atom names should be exported.
        conf_type (ConformerType, optional): Defaults to ConformerType.Ideal.

    Returns:
        str: String representation of the component in the PDB format.
    """
    (mol_to_save, conf_id, conf_type) = _prepate_structure(
        component, remove_hs, conf_type
    )

    info = rdkit.Chem.rdchem.AtomPDBResidueInfo()
    info.SetResidueName(f"{component.id:>3}")
    info.SetTempFactor(20.0)
    info.SetOccupancy(1.0)
    info.SetChainId("A")
    info.SetResidueNumber(1)
    info.SetIsHeteroAtom(True)

    for atom in mol_to_save.GetAtoms():
        flag = _get_alt_atom_name(atom) if alt_names else _get_atom_name(atom)
        atom_name = f"{flag:<4}"  # make sure it is 4 characters
        info.SetName(atom_name)
        atom.SetMonomerInfo(info)

    pdb_title = [
        f"HEADER    {conf_type.name} coordinates",
        f" for PDB-CCD {component.id}",
        f"COMPND    {component.id}",
        f"AUTHOR    pdbccdutils {pdbeccdutils.__version__}",
        f"AUTHOR    RDKit {rdkit.__version__}",
    ]

    try:
        pdb_body = rdkit.Chem.MolToPDBBlock(mol_to_save, conf_id)
    except Exception:
        pdb_body = _to_pdb_str_fallback(
            mol_to_save, component.id, conf_id, conf_type.name
        )

    return "\n".join(pdb_title + [pdb_body])


def to_sdf_str(
    component: Component,
    remove_hs: bool = True,
    conf_type: ConformerType = ConformerType.Ideal,
):
    """Converts structure to the SDF format.

    Args:
        component (Component): Component to be exported.
        remove_hs (bool, optional): Defaults to True.
        conf_type (ConformerType, optional): Defaults to ConformerType.Ideal.

    Raises:
        CCDUtilsError: In case the structure could not be exported.

    Returns:
        str: String representation of the component in the SDF format
    """
    (mol_to_save, _, conf_type) = _prepate_structure(component, remove_hs, conf_type)

    mol_block = []

    if conf_type == ConformerType.AllConformers:
        conformers = [ConformerType.Model, ConformerType.Ideal, ConformerType.Computed]
    else:
        conformers = [conf_type]
    try:
        for c in conformers:
            try:
                conf_id = -1
                if c != ConformerType.AllConformers:
                    conf_id = component.get_conformer(c).GetId()

                block = [
                    f"{component.id} - {c.name} conformer",
                    rdkit.Chem.MolToMolBlock(mol_to_save, confId=conf_id).strip(),
                    "$$$$\n",
                ]
                mol_block += block
            except ValueError as e:
                if str(e) == "Bad Conformer Id":
                    pass
                else:
                    raise CCDUtilsError(f"Error writing SDF file - {e}")
    except Exception:
        mol_block = _to_sdf_str_fallback(mol_to_save, component.id, conformers)

    return "\n".join(mol_block)


def to_xyz_str(component, remove_hs=True, conf_type=ConformerType.Ideal):
    """Converts structure to the XYZ format. Does not yet support
    ConformerType.AllConformers.

    Args:
        component (Component): Component to be exported.
        remove_hs (bool, optional): Defaults to True.
        conf_type (ConformerType, optional): Defaults to ConformerType.Ideal.

    Returns:
        str: String representation of the component in the XYZ format
    """
    (mol_to_save, conf_id, conf_type) = _prepate_structure(
        component, remove_hs, conf_type
    )
    conformer = mol_to_save.GetConformer(id=conf_id)

    result = list()
    result.append(str(mol_to_save.GetNumAtoms()))
    result.append(component.id)

    for atom in mol_to_save.GetAtoms():
        coords = conformer.GetAtomPosition(atom.GetIdx())
        result.append(
            f"{atom.GetSymbol():<4}{coords.x: f}  {coords.y: f}  {coords.z: f}"
        )

    return "\n".join(result)


def to_xml_xml(component, remove_hs=True, conf_type=ConformerType.Ideal):
    """Converts structure to the XML format and returns its XML repr.

    Args:
        component (Component): Component to be exported.
        remove_hs (bool, optional): Defaults to True.
        conf_type (ConformerType, optional): Defaults to ConformerType.Ideal.

    Returns:
        xml.etree.ElementTree.Element: XML object
    """
    root = Element("chemComp")

    id_e = SubElement(root, "id")
    name_e = SubElement(root, "name")
    formula_e = SubElement(root, "formula")
    sys_name_e = SubElement(root, "systematicName")
    s_smiles_e = SubElement(root, "stereoSmiles")
    n_smiles_e = SubElement(root, "nonStereoSmiles")
    inchi_e = SubElement(root, "inchi")

    name_e.text = component.name
    id_e.text = component.id
    formula_e.text = component.formula
    sys_name_e.text = next(
        (
            x.value
            for x in component.descriptors
            if x.type == "SYSTEMATIC NAME" and x.program == "ACDLabs"
        ),
        "",
    )
    s_smiles_e.text = next(
        (
            x.value
            for x in component.descriptors
            if x.type == "SMILES_CANONICAL" and x.program == "CACTVS"
        ),
        "",
    )
    n_smiles_e.text = next(
        (
            x.value
            for x in component.descriptors
            if x.type == "SMILES" and x.program == "CACTVS"
        ),
        "",
    )
    inchi_e.text = component.inchi

    return root


def to_xml_str(component: Component, remove_hs=True, conf_type=ConformerType.Ideal):
    """Converts structure to the XML format. Presently just molecule
    metadata are serialized without any coordinates, which is in
    accordance with the content of the PDBeChem area.

    Args:
        component (Component): Component to be exported.
        remove_hs (bool, optional): Defaults to True.
        conf_type (ConformerType, optional): Defaults to ConformerType.Ideal.

    Returns:
        str: String representation of the component in CML format.
    """
    root = to_xml_xml(component, remove_hs, conf_type)

    xml = ET.tostring(root, encoding="utf-8", method="xml")
    xmls_tring = minidom.parseString(xml)

    return xmls_tring.toprettyxml(indent="  ")


def to_pdb_ccd_cif_file(path, component: Component, remove_hs=True):
    """Converts structure to the PDB CIF format. Both model and ideal
    coordinates are stored. In case ideal coordinates are missing, rdkit
    attempts to generate 3D coordinates of the conformer.

    Args:
        path (str): Path to save cif file.
        component (Component): Component to be exported.
        remove_hs (bool, optional): Defaults to True.
    """

    def remove_hydrogens(cif_block_copy):
        cif_tools.preprocess_cif_category(cif_block_copy, "_chem_comp_atom.")
        cif_tools.preprocess_cif_category(cif_block_copy, "_chem_comp_bond.")

        # scrap hydrogen atoms
        h_names: List[str] = []
        atom_table = cif_block_copy.find("_chem_comp_atom.", ["type_symbol", "atom_id"])
        for i in range(len(atom_table) - 1, -1, -1):
            if atom_table[i][0] == "H":
                h_names.append(atom_table[i][1])
                del atom_table[i]

        # scrap bonds to hydrogen atoms
        if "_chem_comp_bond." not in cif_block_copy.get_mmcif_category_names():
            return

        bond_table = cif_block_copy.find("_chem_comp_bond.", ["atom_id_1", "atom_id_2"])
        for j in range(len(bond_table) - 1, -1, -1):
            if (bond_table[j][0] in h_names) or (bond_table[j][1] in h_names):
                del bond_table[j]

    if not isinstance(component.ccd_cif_block, gemmi.cif.Block):
        component.ccd_cif_block = _to_pdb_ccd_cif_block(component)

    temp_doc = cif.Document()
    cif_block_copy = temp_doc.add_copied_block(component.ccd_cif_block)

    _add_sw_info_cif(cif_block_copy)
    _add_2d_depiction_cif(component, cif_block_copy)
    _add_fragments_and_scaffolds_cif(component, cif_block_copy)
    _add_rdkit_properties_cif(component, cif_block_copy)
    _add_unichem_mapping_cif(component, cif_block_copy)
    _add_rdkit_conformer_cif(component, cif_block_copy, remove_hs)

    if remove_hs:
        remove_hydrogens(cif_block_copy)

    temp_doc.write_file(path, cif.Style.Pdbx)


def to_cml_str(component: Component, remove_hs=True, conf_type=ConformerType.Ideal):
    """Converts structure to the EBI representation of the molecule in
    CML format: http://cml.sourceforge.net/schema/cmlCore.xsd

    Args:
        component (Component): Component to be exported.
        remove_hs (bool, optional): Defaults to True.
        conf_type (ConformerType, optional): Defaults to ConformerType.Ideal.

    Returns:
        str: String representation of the component in CML format.
    """
    (mol_to_save, conf_id, conf_type) = _prepate_structure(
        component, remove_hs, conf_type
    )

    root = ET.Element("cml")
    root.set("xsi:schemaLocation", "http://cml.sourceforge.net/schema/cmlCore.xsd")
    root.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
    root.set("dictRef", "ebiMolecule:ebiMoleculeDict.cml")
    root.set("ebiMolecule", "http://www.ebi.ac.uk/felics/molecule")

    f_charge = sum([a.GetFormalCharge() for a in mol_to_save.GetAtoms()])
    mol = ET.SubElement(
        root, "molecule", {"id": component.id, "formalCharge": str(f_charge)}
    )

    id_inchi = ET.SubElement(mol, "identifier", {"dictRef": "ebiMolecule:inchi"})
    id_inchi.text = component.inchi
    id_systematic = ET.SubElement(
        mol, "identifier", {"dictRef": "ebiMolecule:systematicName"}
    )
    id_systematic.text = component.name
    id_formula1 = ET.SubElement(mol, "formula", {"dictRef": "ebiMolecule:stereoSmiles"})
    id_formula2 = ET.SubElement(
        mol, "formula", {"dictRef": "ebiMolecule:nonStereoSmiles"}
    )
    id_formula1.text = next(
        (
            x.value
            for x in component.descriptors
            if x.type == "SMILES_CANONICAL" and x.program == "CACTVS"
        ),
        "",
    )
    id_formula2.text = next(
        (
            x.value
            for x in component.descriptors
            if x.type == "SMILES" and x.program == "CACTVS"
        ),
        "",
    )

    atom_array = ET.SubElement(mol, "atomArray")
    conformer = mol_to_save.GetConformer(id=conf_id)

    for atom in mol_to_save.GetAtoms():
        element = atom.GetSymbol()
        a_name = _get_atom_name(atom)
        coords = conformer.GetAtomPosition(atom.GetIdx())

        a_entry = ET.SubElement(
            atom_array, "atom", {"id": a_name, "elementType": element}
        )
        a_entry.set("x3", str(coords.x))
        a_entry.set("y3", str(coords.y))
        a_entry.set("z3", str(coords.z))

    bond_array = ET.SubElement(mol, "bondArray")
    for bond in mol_to_save.GetBonds():
        atom_1 = _get_atom_name(bond.GetBeginAtom())
        atom_2 = _get_atom_name(bond.GetEndAtom())
        bond_order = _get_cml_bond_type(bond.GetBondType())

        bond_entry = ET.SubElement(bond_array, "bond")
        bond_entry.set("atomsRefs2", atom_1 + " " + atom_2)
        bond_entry.set("order", str(bond_order))

    cml = ET.tostring(root, encoding="utf-8", method="xml")
    pretty = minidom.parseString(cml)

    return pretty.toprettyxml(indent="  ")


def to_json_dict(component: Component, remove_hs=True, conf_type=ConformerType.Ideal):
    """Returns component information in dictionary suitable for json
    formating

    Args:
        component (Component): Component to be
            exported.
        remove_hs (bool, optional): Defaults to True.
        conf_type (ConformerType, optional):
            Defaults to ConformerType.Ideal.

    Raises:
        AttributeError: If all conformers are requested. This feature is
        not supported not is planned.

    Returns:
        :obj:`dict` of :obj:`str`: dictionary representation of the component
    """
    if conf_type == ConformerType.AllConformers:
        raise AttributeError("All conformer export is not supported for json export")

    (mol_to_save, conf_id, conf_type) = _prepate_structure(
        component, remove_hs, conf_type
    )
    result = {"atoms": [], "bonds": []}
    conformer = mol_to_save.GetConformer(conf_id)

    for atom in mol_to_save.GetAtoms():
        atom_dict = {}

        coords = conformer.GetAtomPosition(atom.GetIdx())

        atom_dict["id"] = atom.GetProp("name")
        atom_dict["element"] = atom.GetSymbol()
        atom_dict["charge"] = atom.GetFormalCharge()

        if conformer.Is3D():
            atom_dict["coords"] = {
                "X": round(coords.x, 4),
                "Y": round(coords.y, 4),
                "Z": round(coords.z, 4),
            }
        else:
            atom_dict["coords"] = {"X": round(coords.x, 4), "Y": round(coords.y, 4)}

        result["atoms"].append(atom_dict)

    for bond in mol_to_save.GetBonds():
        bond_dict = {}
        bond_dict["from"] = bond.GetBeginAtomIdx()
        bond_dict["to"] = bond.GetEndAtomIdx()
        bond_dict["type"] = bond.GetBondType().name

        result["bonds"].append(bond_dict)

    return {component.id: result}


def to_json_str(component: Component, remove_hs=True, conf_type=ConformerType.Ideal):
    """Converts structure into JSON representation. https://www.json.org/

    Args:
        component (Component): Component to be exported.
        remove_hs (bool, optional): Defaults to True.
        conf_type (ConformerType, optional): Defaults to ConformerType.Ideal.

    Returns:
        str: json representation of the component as a string.

    """

    temp = to_json_dict(component, remove_hs, conf_type)
    temp["rdkit_version"] = rdkit.__version__
    temp["pdbeccdutils_version"] = pdbeccdutils.__version__

    return json.dumps(temp)


def _prepate_structure(component, remove_hs, conf_type):
    """Prepare structure for export based on parameters. If deemed
    conformation is missing, an exception is thrown.
    TODO: handling AllConformers for other than PDB, SDF formats.

    Args:
        component (pdbeccdutils.core.Component): Component to be
            exported.
        remove_hs (bool, optional): Defaults to True.
        conf_type (ConformerType, optional):
            Defaults to ConformerType.Ideal.

    Returns:
        tuple(rdkit.Mol,int,ConformerType): mol along with properties
        to be exported.
    """
    conf_id = -1  # this is AllConformers options

    if conf_type == ConformerType.Depiction:
        conf_id = 0
    else:
        for c in component.mol.GetConformers():
            if c.GetProp("name") == conf_type.name:
                conf_id = c.GetId()
                break

    mol_to_save = (
        component.mol2D if conf_type == ConformerType.Depiction else component.mol
    )

    if remove_hs:
        mol_to_save = rdkit.Chem.RemoveHs(mol_to_save, sanitize=False)
        mol_to_save.UpdatePropertyCache(strict=False)

    return (mol_to_save, conf_id, conf_type)


def _to_pdb_ccd_cif_block(component):
    """Export component to the PDB CCD CIF file.

    Args:
        component (pdbeccdutils.core.Component): Component to be
            exported.

    Returns:
        (dict of str: str)): dictionary representation of the component
            to be serialized as a cif file.
    """

    doc = cif.Document()
    cif_block = doc.add_new_block(component.id)

    _write_pdb_ccd_cif_info(cif_block, component)
    _write_pdb_ccd_cif_atoms(cif_block, component)
    _write_pdb_ccd_cif_bonds(cif_block, component)
    _write_pdb_ccd_cif_descriptor(cif_block, component)

    return cif_block


def _write_pdb_ccd_cif_info(cif_block, component):
    """Writes _chem_comp namespace with general information about the
    component

    Args:
        cif_block (cif.Block): mmcif Block object from gemmi.
        component (Component): Component to be exported.
    """

    calc_formula = rdkit.Chem.rdMolDescriptors.CalcMolFormula(component.mol)
    calc_weight = rdkit.Chem.rdMolDescriptors.CalcExactMolWt(component.mol)
    date = component.modified_date
    mod_date = f"{date.year}-{date.month:02d}-{date.day:02d}"

    label = "_chem_comp."
    cif_block.set_pairs(
        label,
        {
            "id": component.id,
            "type": "NON-POLYMER",
            "pdbx_type": "HETAIN",
            "formula": cif.quote(component.formula)
            if component.formula
            else cif.quote(calc_formula),
            "formula_weight": f"{calc_weight:.3f}",
            "three_letter_code": component.id,
            "pdbx_type": "HETAIN",
            "pdbx_modified_date": mod_date,
            "pdbx_release_status": component.pdbx_release_status.name,
        },
        raw=True,
    )


def _write_pdb_ccd_cif_atoms(cif_block, component):
    """Writes the _chem_comp_atom namespace with atom details.
    Controlled dictionary:
    http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Categories/chem_comp_atom.html

    Args:
        cif_block (cif.Block): mmcif Block object from gemmi.
        component (Component): Component to be exported.
    """

    if component.mol.GetNumAtoms() < 1:
        return

    # if we dont have ideal coordinates, we just try to compute them
    ideal_type = ConformerType.Ideal
    if component.has_degenerated_conformer(ConformerType.Ideal):
        ideal_type = (
            ConformerType.Computed if component.compute_3d() else ConformerType.Ideal
        )

    label = "_chem_comp_atom."
    atom_fields = [
        "comp_id",
        "atom_id",
        "alt_atom_id",
        "type_symbol",
        "charge",
        "pdbx_align",
        "pdbx_aromatic_flag",
        "pdbx_leaving_atom_flag",
        "pdbx_stereo_config",
        "model_Cartn_x",
        "model_Cartn_y",
        "model_Cartn_z",
        "pdbx_model_Cartn_x_ideal",
        "pdbx_model_Cartn_y_ideal",
        "pdbx_model_Cartn_z_ideal",
        "pdbx_component_atom_id",
        "pdbx_component_comp_id",
        "pdbx_ordinal",
    ]
    atom_loop = cif_block.init_loop(label, atom_fields)

    for atom in component.mol.GetAtoms():
        at_id = atom.GetIdx()
        model_atom = _get_atom_coord(component, at_id, ConformerType.Model)
        ideal_atom = _get_atom_coord(component, at_id, ideal_type)

        new_row = [
            component.id,
            _get_atom_name(atom),
            _get_alt_atom_name(atom),
            atom.GetSymbol(),
            str(atom.GetFormalCharge()),
            "?",
            "Y" if atom.GetIsAromatic() else "N",
            "N",
            _get_ccd_cif_chiral_type(atom),
            f"{model_atom.x:.3f}",
            f"{model_atom.y:.3f}",
            f"{model_atom.z:.3f}",
            f"{ideal_atom.x:.3f}",
            f"{ideal_atom.y:.3f}",
            f"{ideal_atom.z:.3f}",
            _get_atom_name(atom),
            component.id,
            str(atom.GetIdx() + 1),
        ]

        atom_loop.add_row(cif.quote_list(new_row))


def _write_pdb_ccd_cif_bonds(cif_block, component):
    """Writes the _chem_comp_bond namespace with atom details.
    Controlled dictionary:
    http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Categories/chem_comp_bond.html

    Args:
        cif_block (cif.Block): mmcif Block object from gemmi.
        component (Component): Component to be exported.
    """
    if component.mol.GetNumBonds() < 1:
        return

    label = "_chem_comp_bond."
    bond_fields = [
        "comp_id",
        "atom_id_1",
        "atom_id_2",
        "value_order",
        "pdbx_aromatic_flag",
        "pdbx_stereo_config",
        "pdbx_ordinal",
    ]
    bond_loop = cif_block.init_loop(label, bond_fields)

    for b in component.mol.GetBonds():
        atom_a = b.GetBeginAtom()
        atom_b = b.GetEndAtom()

        new_row = [
            component.id,
            _get_atom_name(atom_a),
            _get_atom_name(atom_b),
            _get_ccd_cif_bond_type(b),
            "Y" if b.GetIsAromatic() else "N",
            _get_ccd_cif_bond_stereo(b),
            str(b.GetIdx() + 1),
        ]
        bond_loop.add_row(cif.quote_list(new_row))


def _write_pdb_ccd_cif_descriptor(cif_block, component):
    """Writes the _pdbx_chem_comp_descriptor namespace with details.

    Args:
        cif_block (cif.Block): mmcif Block object from gemmi.
        component (Component): Component to be exported.
    """

    if not component.descriptors:
        return

    label = "_pdbx_chem_comp_descriptor."

    descriptor_fields = ["comp_id", "type", "program", "descriptor"]
    descriptor_loop = cif_block.init_loop(label, descriptor_fields)

    for entry in component.descriptors:
        new_row = [component.id, entry.type, entry.program, entry.value]
        descriptor_loop.add_row(cif.quote_list(new_row))


def _get_atom_name(atom):
    """Gets atom name. If not set ElementSymbol + Id is used.

    Args:
        atom (rdkit.Chem.rdchem.Atom): rdkit atom.

    Returns:
        str: Name of the atom.
    """
    return (
        atom.GetProp("name")
        if atom.HasProp("name")
        else atom.GetSymbol() + str(atom.GetIdx())
    )


def _get_alt_atom_name(atom):
    """Gets alternate atom name. If not set _get_atom_name method is
    used.

    Args:
        atom (rdkit.Chem.rdchem.Atom): rdkit atom.

    Returns:
        str: Name of the atom.
    """
    return (
        atom.GetProp("alt_name") if atom.HasProp("alt_name") else _get_atom_name(atom)
    )


def _get_cml_bond_type(bond_order):
    """Translate bond type from rdkit to CML language.

    Args:
        bond_order (rdkit.Chem.rdchem.BondType): rdkit bond type

    Returns:
        str: bond type in the CML language.
    """
    if bond_order == rdkit.Chem.rdchem.BondType.SINGLE:
        return "1"

    if bond_order == rdkit.Chem.rdchem.BondType.DOUBLE:
        return "2"

    if bond_order == rdkit.Chem.rdchem.BondType.TRIPLE:
        return "3"

    if bond_order == rdkit.Chem.rdchem.BondType.AROMATIC:
        return "A"
    else:
        return str(bond_order)


def _get_ccd_cif_bond_stereo(bond):
    """Get bond stereochemistry information to be used in CCD CIF file.
    Controlled dictionary from: http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_chem_comp_bond.pdbx_stereo_config.html

    Args:
        bond (rdkit.Chem.rdchem.Bond): Molecular bond.

    Returns:
        str: bond stereochemistry information for the field
            '_chem_comp_bond.pdbx_stereo_config'.
    """

    stereo = bond.GetStereo()

    if stereo in (
        rdkit.Chem.rdchem.BondStereo.STEREOE,
        rdkit.Chem.rdchem.BondStereo.STEREOCIS,
    ):
        return "E"

    if stereo in (
        rdkit.Chem.rdchem.BondStereo.STEREOZ,
        rdkit.Chem.rdchem.BondStereo.STEREOTRANS,
    ):
        return "Z"

    return "N"


def _get_ccd_cif_bond_type(bond):
    """Translate bond type from rdkit to the CCD CIF format.
    Controlled dictionary from:
    http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_chem_comp_bond.value_order.html

    Args:
        bond_order (rdkit.Chem.rdchem.Bond): rdkit molecular bond.

    Returns:
        str: bond type for the '_chem_comp_bond.value_order' field. If
            none of the rdkit bond types can be matched SING is returned.
    """
    bond_order = bond.GetBondType()

    if bond_order == rdkit.Chem.rdchem.BondType.SINGLE:
        return "SING"

    if bond_order == rdkit.Chem.rdchem.BondType.DOUBLE:
        return "DOUB"

    if bond_order == rdkit.Chem.rdchem.BondType.TRIPLE:
        return "TRIP"

    if bond_order == rdkit.Chem.rdchem.BondType.AROMATIC:
        return "AROM"

    if bond_order == rdkit.Chem.rdchem.BondType.QUADRUPLE:
        return "QUAD"

    return "SING"


def _get_ccd_cif_chiral_type(atom):
    """Translate atom chiral from rdkit to the CCD CIF format.
    Controlled dictionary from:
    http://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_chem_comp_atom.pdbx_stereo_config.html

    Args:
        atom (rdkit.Chem.rdchem.Atom): rdkit atom

    Returns:
        str: chiral type for the '_chem_comp_atom.pdbx_stereo_config'
            field. If none of the rdkit atom types can be matched 'N' is
            returned.
    """
    chiral_type = atom.GetChiralTag()

    if chiral_type == rdkit.Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW:
        return "R"

    if chiral_type == rdkit.Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW:
        return "S"

    return "N"


def _get_atom_coord(component, at_id, conformer_type):
    """Retrieve 3D coordinates for the particular atom in a given
    conformer.

    Args:
        component (Component): component to be processed.
        at_id (int): atom id to be retrieved.
        conformer_type (ConformerType): conformer type

    Returns:
        rdkit.Geometry.rdGeometry.Point3D: 3D coordinates of the atom.
    """
    conformer = None
    for c in component.mol.GetConformers():
        if c.GetProp("name") == conformer_type.name:
            conformer = c
            break

    return conformer.GetAtomPosition(at_id)


# region fallbacks


def _to_sdf_str_fallback(mol, ccd_id, conformers):
    """Fallback method to generate SDF file in case the default one in
    RDKit fails.

    Args:
        mol (rdkit.Chem.rdchem.Mol): rdkit mol to be exported
        id (str): component id.
        conformers (list of of ConformerType): List of conformer types
            to be exported.

    Returns:
        list of str: SDF representation of the component
    """
    content = []

    for c in conformers:
        rdkit_conformer = None

        if c == ConformerType.AllConformers:
            rdkit_conformer = mol.GetConformer()
        else:
            for conf in mol.GetConformers():
                if conf.GetProp("name") == c.name:
                    rdkit_conformer = conf
                    break

        atom_count = mol.GetNumAtoms()
        bond_count = mol.GetNumBonds()

        content += [
            f"{ccd_id} - {c.name} conformer",
            "    RDKit   3D",
            "\n" f"{atom_count:>3}{bond_count:3}  0  0  0  0  0  0  0  0999 V2000",
        ]

        for i in range(0, atom_count):
            pos = rdkit_conformer.GetAtomPosition(i)
            atom = mol.GetAtomWithIdx(i)
            sdf_charge = __charge_to_sdf(atom.GetFormalCharge())
            content.append(
                f"{pos.x:>10.4f}{pos.y:>10.4f}{pos.z:>10.4f} {atom.GetSymbol():<3} 0{sdf_charge:>3}"
            )

        for i in range(0, bond_count):
            bond = mol.GetBondWithIdx(i)
            content.append(
                "{:>3}{:>3}{:>3}{:>3}  0  0  0".format(
                    bond.GetBeginAtom().GetIdx() + 1,
                    bond.GetEndAtom().GetIdx() + 1,
                    __bond_type_to_sdf(bond),
                    __bond_stereo_to_sdf(bond),
                )
            )
        content.append("M  END")
        content.append("$$$$\n")
    return content


def _to_pdb_str_fallback(mol, component_id, conf_id, conf_name="Model"):
    """Fallback method to generate PDB file in case the default one in
    RDKit fails.

    Args:
        mol (rdkit.Chem.rdchem.Mol): Molecule to be written.
        component_id (str): Component id.
        conf_id (int): conformer id to be written.
        conf_name (str): conformer name to be written.

    Returns:
        str: String representation the component in the PDB format.
    """
    conformer_ids = []
    content = [
        f"HEADER    {conf_name} coordinates for PDB-CCD {component_id}",
        f"COMPND    {component_id}",
        f"AUTHOR    pdbccdutils {pdbeccdutils.__version__}",
        f"AUTHOR    RDKit {rdkit.__version__}",
    ]

    if conf_id == -1:
        conformer_ids = [c.GetId() for c in mol.GetConformers()]
    else:
        conformer_ids = [conf_id]

    for m in conformer_ids:
        rdkit_conformer = mol.GetConformer(m)

        for i in range(0, mol.GetNumAtoms()):
            atom = mol.GetAtomWithIdx(i)

            s = "{:<6}{:>5} {:<4} {:>3} {}{:>4}{}   {:>8.3f}{:>8.3f}{:>8.3f}{:>6.2f}{:>6.2f}          {:>2}{:>2}".format(
                "HETATM",
                i + 1,
                atom.GetProp("name"),
                component_id,
                "A",
                1,
                " ",
                rdkit_conformer.GetAtomPosition(i).x,
                rdkit_conformer.GetAtomPosition(i).y,
                rdkit_conformer.GetAtomPosition(i).z,
                1,
                20,
                atom.GetSymbol(),
                atom.GetFormalCharge(),
            )
            content.append(s)

        for i in range(0, mol.GetNumAtoms()):
            pivot = mol.GetAtomWithIdx(i)
            s = "CONECT{:>5}".format(i + 1)

            for b in pivot.GetBonds():
                end_atom = b.GetOtherAtomIdx(i)

                if end_atom < i:
                    continue

                for t in range(0, math.floor(b.GetBondTypeAsDouble())):
                    s += "{:>5}".format(end_atom + 1)

            if len(s) > 11:
                content.append(s)

        content.append("END")

    return "\n".join(content)


# endregion fallbacks


# region helpers
def __charge_to_sdf(charge):
    """Translate RDkit charge to the SDF language.

    Args:
        charge (int): Numerical atom charge.

    Returns:
        str: Str representation of a charge in the sdf language
    """
    if charge == -3:
        return "7"
    if charge == -2:
        return "6"
    if charge == -1:
        return "5"
    if charge == 0:
        return "0"
    if charge == 1:
        return "+1"
    if charge == 2:
        return "+2"
    if charge == 3:
        return "+4"

    return "0"


def __bond_stereo_to_sdf(bond):
    """Translate bond stereo information to the sdf language. Needs to
    be checked.

    Args:
        bond (rdkit.Chem.rdchem.Bond): bond to be processed.

    Returns:
        str: bond type in sdf language.
    """
    stereo = bond.GetStereo()

    if stereo == rdkit.Chem.rdchem.BondStereo.STEREONONE:
        return "0"
    if stereo == rdkit.Chem.rdchem.BondStereo.STEREOANY:
        return "4"
    if stereo in (
        rdkit.Chem.rdchem.BondStereo.STEREOCIS,
        rdkit.Chem.rdchem.BondStereo.STEREOTRANS,
    ):
        return "3"
    return "0"


def __bond_type_to_sdf(bond):
    """Get bond type in sdf language. Based on:
    http://www.nonlinear.com/progenesis/sdf-studio/v0.9/faq/sdf-file-format-guidance.aspx

    Args:
        bond_type (rdkit.Chem.rdchem.Bond): Bond to be processed

    Returns:
        str: String representation of the bond type.
    """
    bond_order = bond.GetBondType()

    if bond_order == rdkit.Chem.rdchem.BondType.SINGLE:
        return "1"
    if bond_order == rdkit.Chem.rdchem.BondType.DOUBLE:
        return "2"
    if bond_order == rdkit.Chem.rdchem.BondType.TRIPLE:
        return "3"
    if bond_order == rdkit.Chem.rdchem.BondType.AROMATIC:
        return "A"

    return "0"


# endregion helpers


# region cif-export addition


def _add_sw_info_cif(cif_block_copy):
    """Add information to the cif file about software versions used
    to generate this information

    Args:
        cif_block_copy (Block: gemmi.cif.Block): Block representation of the molecule
        from gemmi.
    """
    category = "_software."
    sw_fields = ["name", "version", "description"]
    sw_loop = cif_block_copy.init_loop(category, sw_fields)
    sw_loop.add_row(
        cif.quote_list(["rdkit", rdkit.__version__, cif.quote("Core functionality.")])
    )
    sw_loop.add_row(
        cif.quote_list(
            [
                "pdbeccdutils",
                pdbeccdutils.__version__,
                cif.quote("Wrapper to provide 2D templates and molecular fragments."),
            ]
        )
    )


def _add_2d_depiction_cif(component, cif_block_copy):
    """Add 2D coordinates of the component depiction

    Args:
        component (Component): pdbeccdutils component.
        cif_block_copy (Block: gemmi.cif.Block): Block representation of the molecule
        from gemmi.
    """
    if component.mol2D is None:
        return

    __add_rdkit_2d_atoms_cif(component, cif_block_copy)
    __add_rdkit_2d_bonds_cif(component, cif_block_copy)


def _add_fragments_and_scaffolds_cif(component, cif_block_copy):
    """Add fragments and scaffolds information to the CIF export

    Args:
        component (Component): Component to be exported.
        cif_block_copy (Block: gemmi.cif.Block): Block representation of the molecule
        from gemmi.
    """
    substructure_category = "_pdbe_chem_comp_substructure."
    substructure_fields = [
        "comp_id",
        "substructure_name",
        "id",
        "substructure_type",
        "substructure_smiles",
        "substructure_inchis",
        "substructure_inchikeys",
    ]
    substructure_loop = cif_block_copy.init_loop(
        substructure_category, substructure_fields
    )

    for i, scaffold in enumerate(component.scaffolds):
        mol = rdkit.Chem.MolFromSmiles(scaffold.smiles)
        new_row = [
            component.id,
            scaffold.name,
            f"S{i+1}",
            "scaffold",
            scaffold.smiles,
            rdkit.Chem.MolToInchi(mol),
            rdkit.Chem.MolToInchiKey(mol),
        ]
        substructure_loop.add_row(cif.quote_list(new_row))

    for j, fragment in enumerate(component.fragments):
        mol = rdkit.Chem.MolFromSmiles(fragment.smiles)
        new_row = [
            component.id,
            fragment.name,
            f"F{j+1}",
            "fragment",
            fragment.smiles,
            rdkit.Chem.MolToInchi(mol),
            rdkit.Chem.MolToInchiKey(mol),
        ]
        substructure_loop.add_row(cif.quote_list(new_row))

    mapping_category = "_pdbe_chem_comp_substructure_mapping."
    mapping_fields = ["comp_id", "atom_id", "substructure_id", "substructure_ordinal"]
    mapping_loop = cif_block_copy.init_loop(mapping_category, mapping_fields)

    for i, scaffold in enumerate(component.scaffolds):
        for a, mapping in enumerate(scaffold.mappings):
            for atom_name in mapping:
                new_row = [component.id, atom_name, f"S{i+1}", f"{a+1}"]
                mapping_loop.add_row(cif.quote_list(new_row))

    for i, fragment in enumerate(component.fragments):
        for a, mapping in enumerate(fragment.mappings):
            for atom_name in mapping:
                new_row = [component.id, atom_name, f"F{i+1}", f"{a+1}"]
                mapping_loop.add_row(cif.quote_list(new_row))


def _add_rdkit_properties_cif(component, cif_block_copy):
    """Add properties calculated by the rdkit to the dictionary

    Args:
        component (Component): pdbeccdutils component.
        cif_block (Block: gemmi.cif.Block): Block representation of the molecule
        from gemmi.
    """
    category = "_pdbe_chem_comp_rdkit_properties."
    cif_block_copy.set_pairs(category, {"comp_id": component.id})

    for k, v in component.physchem_properties.items():
        cif_block_copy.set_pairs(
            category, {k: f"{v:.0f}" if v.is_integer() else f"{v:.3f}"}
        )


def __add_rdkit_2d_atoms_cif(component, cif_block_copy):
    """Add atom coordinates to the mmCIF file.

    Args:
        component (Component): pdbeccdutils component.
        cif_block_copy (Block: gemmi.cif.Block): Block representation of the molecule
        from gemmi.
    """
    category = "_pdbe_chem_comp_atom_depiction."
    atom_depiction_fields = [
        "comp_id",
        "atom_id",
        "element",
        "model_Cartn_x",
        "model_Cartn_y",
        "pdbx_ordinal",
    ]
    conformer = component.mol2D.GetConformer()
    atom_depiction_loop = cif_block_copy.init_loop(category, atom_depiction_fields)

    for i, atom in enumerate(component.mol2D.GetAtoms()):
        new_row = [
            component.id,
            atom.GetProp("name"),
            atom.GetSymbol(),
            f"{conformer.GetAtomPosition(i).x:.3f}",
            f"{conformer.GetAtomPosition(i).y:.3f}",
            str(i + 1),
        ]
        atom_depiction_loop.add_row(cif.quote_list(new_row))


def __add_rdkit_2d_bonds_cif(component, cif_block_copy):
    """Add bond information to the mmCIF file.

    Args:
        component (Component): pdbeccdutils component.
        cif_block (Block: gemmi.cif.Block): Block representation of the molecule
        from gemmi.
    """
    category = "_pdbe_chem_comp_bond_depiction."
    fields = [
        "comp_id",
        "atom_id_1",
        "atom_id_2",
        "value_order",
        "bond_dir",
        "pdbx_ordinal",
    ]
    bonds_depiction_loop = cif_block_copy.init_loop(category, fields)

    try:
        copy = rdkit.Chem.Draw.rdMolDraw2D.PrepareMolForDrawing(
            component.mol2D, wedgeBonds=True, kekulize=True, addChiralHs=True
        )
    except (RuntimeError, ValueError):
        try:
            copy = rdkit.Chem.Draw.rdMolDraw2D.PrepareMolForDrawing(
                component.mol2D, wedgeBonds=False, kekulize=True, addChiralHs=True
            )
        except (RuntimeError, ValueError):
            copy = rdkit.Chem.Draw.rdMolDraw2D.PrepareMolForDrawing(
                component.mol2D, wedgeBonds=False, kekulize=True, addChiralHs=False
            )

    bond_depiction_ordinal = 0
    for i, b in enumerate(copy.GetBonds()):
        if b.GetEndAtom().GetSymbol() != "H":
            bond_depiction_ordinal += 1
            new_row = [
                component.id,
                b.GetBeginAtom().GetProp("name"),
                b.GetEndAtom().GetProp("name"),
                b.GetBondType().name,
                b.GetBondDir().name,
                str(bond_depiction_ordinal),
            ]
            bonds_depiction_loop.add_row(cif.quote_list(new_row))


def _add_unichem_mapping_cif(component, cif_block_copy):
    """Add UniChem mapping to CCD CIF.

    Args:
        component (Component): pdbeccdutils component.
        cif_block_copy (Block: gemmi.cif.Block): Block representation of the molecule
        from gemmi.
    """

    category = "_pdbe_chem_comp_external_mappings."
    fields = ["comp_id", "source", "resource", "resource_id"]
    external_mapping_loop = cif_block_copy.init_loop(category, fields)
    for mapping in component.external_mappings:
        new_row = [component.id, "UniChem", mapping[0], mapping[1]]
        external_mapping_loop.add_row(cif.quote_list(new_row))


def _add_rdkit_conformer_cif(component, cif_block_copy, remove_hs):
    """Add 3D coordinates generated by RDKit.

    component (Component): pdbeccdutils component.
    cif_block (Block: gemmi.cif.Block): Block representation of the molecule
        from gemmi.
    remove_hs (boolean): Whether or not hydrogen atoms should be written
    """

    try:
        conformer = component.get_conformer(ConformerType.Computed)
    except ValueError:
        logging.warning(f"Computed conformer for {component.id} does not exist.")
        return  # no conformer nothing to write, we quit

    category = "_pdbe_chem_comp_rdkit_conformer."
    rdkit_conformer_fields = [
        "comp_id",
        "atom_id",
        "Cartn_x_rdkit",
        "Cartn_y_rdkit",
        "Cartn_z_rdkit",
        "rdkit_method",
        "rdkit_ordinal",
    ]
    rdkit_conformer_loop = cif_block_copy.init_loop(category, rdkit_conformer_fields)
    method = conformer.GetProp("coord_generation")

    # we need to get all the allowed incides
    if remove_hs:
        atom_indices = [
            i
            for i in range(0, component.mol.GetNumAtoms())
            if component.mol.GetAtomWithIdx(i).GetSymbol() != "H"
        ]
    else:
        atom_indices = list(range(0, component.mol.GetNumAtoms()))

    for i, atom_index in enumerate(atom_indices):
        new_row = [
            component.id,
            component.mol.GetAtomWithIdx(atom_index).GetProp("name"),
            f"{conformer.GetAtomPosition(atom_index).x:.3f}",
            f"{conformer.GetAtomPosition(atom_index).y:.3f}",
            f"{conformer.GetAtomPosition(atom_index).z:.3f}",
            method,
            str(i + 1),
        ]
        rdkit_conformer_loop.add_row(cif.quote_list(new_row))


# endregion



#!/usr/bin/env python
# software from PDBe: Protein Data Bank in Europe; https://pdbe.org
#
# Copyright 2018 EMBL - European Bioinformatics Institute
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

"""Module helper providing drawing functionality based on RDKit
"""

import os
import re
import xml.etree.ElementTree as ET
from collections import OrderedDict
from sys import platform

import rdkit
from PIL import Image, ImageDraw, ImageFont

MIN_IMG_DIMENSION = 300
svg_namespace = {"svg": "http://www.w3.org/2000/svg"}


def save_no_image(path_to_image, default_msg=None, width=200):
    """
    Generate pretty image with 'No image available' message in case
    the 2D depiction cannot be created.

    Args:
        path_to_image (str): path to the image
        width (int, optional): Defaults to 200. width of the image
    """
    if path_to_image.split(".")[-1] == "svg":
        svg = (
            _svg_no_image_with_id(default_msg, width)
            if default_msg
            else _svg_no_image(width)
        )
        with open(path_to_image, "w") as f:
            f.write(svg)
    else:
        _png_no_image(path_to_image, width)


def draw_molecule(mol, drawer, file_name, wedge_bonds, atom_highlight, bond_highlight):
    """Draw SVG image from the RDKit molecule.

    Args:
        mol (rdkit.Chem.rdchem.Mol): Rdkit mol object to be depicted.
        drawer (rdkit.Chem.Draw.MolDrawing.DrawingOptions): RDKit object
            with parameters for drawing depiction.
        file_name (str): Path where the depiction will be saved.
        wedge_bonds (bool): Whether or not to wedge bonds.
        atom_highlight (dict): Dictionary with atom id and RGB
            mapping color mapping.
        bond_highlight (dict): Dictionary with mapping of atom
            ids and RGB colors.
    """
    try:
        copy = rdkit.Chem.Draw.rdMolDraw2D.PrepareMolForDrawing(
            mol, wedgeBonds=wedge_bonds, kekulize=True, addChiralHs=True
        )
    except (RuntimeError, ValueError):
        try:
            copy = rdkit.Chem.Draw.rdMolDraw2D.PrepareMolForDrawing(
                mol, wedgeBonds=False, kekulize=True, addChiralHs=True
            )
        except (RuntimeError, ValueError):
            copy = rdkit.Chem.Draw.rdMolDraw2D.PrepareMolForDrawing(
                mol, wedgeBonds=False, kekulize=True, addChiralHs=False
            )

    if bond_highlight is None:
        drawer.DrawMolecule(
            copy,
            highlightAtoms=atom_highlight.keys(),
            highlightAtomColors=atom_highlight,
        )
    else:
        drawer.DrawMolecule(
            copy,
            highlightAtoms=atom_highlight.keys(),
            highlightAtomColors=atom_highlight,
            highlightBonds=bond_highlight.keys(),
            highlightBondColors=bond_highlight,
        )
    drawer.FinishDrawing()

    with open(file_name, "w") as f:
        svg = drawer.GetDrawingText()
        f.write(svg)


def get_drawing_scale(mol):
    """Calculate molecule resolution given 50points per Angstroom

    Args:
        mol (rdkit.Chem.rdchem.Mol): Rdkit mol object.

    Returns:
        [tuple[int,int]]: Dimension of the depictions (x, y).
    """
    cnf = mol.GetConformer()

    a = rdkit.Geometry.Point2D(0, 0)
    b = rdkit.Geometry.Point2D(0, 0)
    a.x = b.x = cnf.GetAtomPosition(0).x
    a.y = b.y = cnf.GetAtomPosition(0).y

    for i in range(1, cnf.GetNumAtoms()):
        a.x = min(a.x, cnf.GetAtomPosition(i).x)
        a.y = min(a.y, cnf.GetAtomPosition(i).y)
        b.x = max(b.x, cnf.GetAtomPosition(i).x)
        b.y = max(b.y, cnf.GetAtomPosition(i).y)

    w = int(50 * (b.x - a.x) + 1)
    h = int(50 * (b.y - a.y) + 1)

    return (max(w, MIN_IMG_DIMENSION), max(h, MIN_IMG_DIMENSION))


def convert_svg(svg_string, ccd_id, mol: rdkit.Chem.Mol):
    """Parse information from SVG depiction into object.

    Args:
        svg_string (str): SVG as string.
        ccd_id (str): CCD ID.
        mol (rdkit.Chem.Mol): RDKit mol object used for depiction.

    Returns:
        dict: JSON-style information for the 2D depiction.
    """

    result_bag = OrderedDict(
        [("ccd_id", ccd_id), ("resolution", {}), ("atoms", []), ("bonds", [])]
    )
    svg = ET.fromstring(svg_string)

    atom_elems = svg.findall("svg:circle", svg_namespace)
    path_elems = svg.findall("svg:path", svg_namespace)
    dimensions_svg = svg.find("svg:rect", svg_namespace)

    atoms = _parse_atoms_from_svg(atom_elems, mol)
    bonds = _parse_bonds_from_svg(path_elems, mol)
    _parse_labels_from_svg(path_elems, atoms)

    result_bag["atoms"] = atoms
    result_bag["bonds"] = bonds

    result_bag["resolution"] = {
        "x": float(dimensions_svg.attrib.get("width")),
        "y": float(dimensions_svg.attrib.get("height")),
    }

    return result_bag


def _parse_atoms_from_svg(atom_elements, mol: rdkit.Chem.Mol):
    """Extract atoms from the SVG atom elements

    Args:
        atom_elements (list[xml.etree.ElementTree.Element]): List of extracted XML elements
        mol (rdkit.Chem.rdchem.Mol): RDkit molecule

    Returns:
        list[dict]: list of JSON-style atom representation.
    """
    result = []
    for atom_svg in atom_elements:
        try:
            atom_id_str = re.search(r"\d+", atom_svg.attrib.get("class")).group(0)
            atom_id = int(atom_id_str)

            if atom_id >= mol.GetNumAtoms():
                continue

            temp = {
                "name": mol.GetAtomWithIdx(atom_id).GetProp("name"),
                "labels": [],
                "x": float(atom_svg.attrib.get("cx")),
                "y": float(atom_svg.attrib.get("cy")),
            }
            result.append(temp)
        except RuntimeError:
            pass  # we do not care for H atoms

    return result


def _parse_labels_from_svg(path_elements, atoms):
    """Parse atom label information from the SVG.

    Args:
        path_elements (list[xml.etree.ElementTree.Element]):
            List of path elements that encode bonds and text.
        atoms (list[dict]): JSON-style representation of atoms.
    """
    atom_id_re = r"atom-\d+"
    for label_svg in path_elements:
        try:
            match = re.fullmatch(atom_id_re, label_svg.attrib["class"])
            if not match:
                continue

            atom_id = int(match.group(0)[5:])
            atoms[atom_id]["labels"].append(
                {"d": label_svg.attrib["d"], "fill": label_svg.attrib["fill"]}
            )
        except (IndexError, KeyError):
            pass  # we do not care for H labels and radicals


def _parse_bonds_from_svg(bond_elements, mol):
    """Extract bonding information from SVG elements

    Args:
        bond_elements (list[xml.etree.ElementTree.Element]):
            List of SVG path elements.
        mol (rdkit.Chem.rdchem.Mol): RDKit mol object]

    Returns:
        list[dict]: JSON-style formated bond informations
    """
    result = []
    re_bond_regex = r"bond-\d+"
    for bond_svg in bond_elements:
        try:
            if not re.search(re_bond_regex, bond_svg.attrib["class"]):
                continue

            atoms = re.findall(r"atom-\d+", bond_svg.attrib["class"])
            atom_id_a = int(atoms[0][5:])  # to get int part of "atom-0"
            atom_id_b = int(atoms[1][5:])

            temp = {
                "bgn": mol.GetAtomWithIdx(atom_id_a).GetProp("name"),
                "end": mol.GetAtomWithIdx(atom_id_b).GetProp("name"),
                "coords": bond_svg.attrib["d"],
                "style": bond_svg.attrib["style"],
            }

            result.append(temp)
        except (RuntimeError, KeyError):
            pass  # we do not care about bonded Hydrogens
    return result


def _png_no_image(path_to_image, width):
    """
    Save image with the text 'No image available' as a png.

    Args:
        path_to_image (str): path to save the image
        width (int): Width of an image
    """

    font = None
    font_path = _supply_font()

    if font is not None:
        font_path = ImageFont.truetype(font_path, size=(int(width / 8)))
    else:
        font = ImageFont.load_default()

    white = (255, 255, 255)
    black = (0, 0, 0)
    img = Image.new("RGBA", (width, width), white)
    draw = ImageDraw.Draw(img)
    draw.multiline_text(
        (width / 4, width / 3),
        "No image\n available",
        font=font,
        align="center",
        fill=black,
    )
    draw = ImageDraw.Draw(img)
    img.save(path_to_image)


def _svg_no_image(width=200):
    """Get svg representation
        width (int, optional): Defaults to 200. width of the image

    Returns:
        str: string representation of an svg image.
    """

    font = width / 8
    svg = f"""<?xml version='1.0' encoding='iso-8859-1'?>
            <svg version='1.1' baseProfile='full'
              xmlns='http://www.w3.org/2000/svg'
                      xmlns:rdkit='http://www.rdkit.org/xml'
                      xmlns:xlink='http://www.w3.org/1999/xlink'
                  xml:space='preserve' width='{width}px' height='{width}px' >
            <rect style='opacity:1.0;fill:#FFFFFF;stroke:none' width='{width}' height='{width}' x='0' y='0'> </rect>
            <text alignment-baseline="middle" text-anchor="middle" x="25%" y="25%" style='font-size:{font}px;font-family:sans-serif;text-anchor:start;fill:#000000'>
                Image not
            </text>
            <text alignment-baseline="middle" text-anchor="middle" x="30%" y="50%" style='font-size:{font}px;font-family:sans-serif;text-anchor:start;fill:#000000'>
                available
            </text>
            </svg>
          """
    return svg


def _svg_no_image_with_id(name, width=200):
    """Get svg representation
        name (str): Name of the component to be displayed, in case
            the image cannot be generated.
        width (int, optional): Defaults to 200. width of the image

    Returns:
        str: string representation of an svg image.
    """
    font = width / 8
    svg = f"""<?xml version='1.0' encoding='iso-8859-1'?>
            <svg version='1.1' baseProfile='full'
              xmlns='http://www.w3.org/2000/svg'
                      xmlns:rdkit='http://www.rdkit.org/xml'
                      xmlns:xlink='http://www.w3.org/1999/xlink'
                  xml:space='preserve' width='{width}px' height='{width}px' >
            <rect style='opacity:1.0;fill:#FFFFFF;stroke:none' width='{width}' height='{width}' x='0' y='0'> </rect>
            <text  dominant-baseline="middle" text-anchor="middle" x="40%" y="50%" style='font-size:{font}px;font-family:sans-serif;text-anchor:start;fill:#000000'>
                {name}
            </text>
            </svg>
          """
    return svg


def _supply_font():
    """
    Platform non-specific function to locate sans-serif font in the environment.

    Returns:
        str: path to the font
    """
    font = ""
    if platform == "linux" or platform == "linux2":
        font = "/usr/share/fonts/gnu-free/FreeSans.ttf"
    elif platform == "darwin":
        font = "/Library/Fonts/arial.ttf"
    elif platform == "win32":
        font = "c:\\windows\\font\\arial.ttf"

    if os.path.isfile(font):
        return font

    return None



"""Test fragment library functionality
"""

import os


def test__img_crated(library, tmpdir):
    file_path = str(tmpdir.join("library.svg"))

    library.to_image(file_path)

    assert os.path.isfile(file_path)


def test_img_png_created(library, tmpdir):
    file_path = str(tmpdir.join("library.png"))
    library.to_image(file_path, source="PDBe")

    assert os.path.isfile(file_path)


def test_generate_conformers(library):
    library.generate_conformers()

    for entry in library.library.values():
        assert entry.mol.GetConformers()



"""
Unittest of the process_components_cif_cli.py command line script

method based on
http://dustinrcollins.com/testing-python-command-line-apps
adapted to use nose then converted to pytest
"""
import json
import os
import re
import xml.etree.ElementTree as ET
from pathlib import Path

import pytest
from pdbeccdutils.helpers.drawing import svg_namespace
from pdbeccdutils.scripts.process_components_cif_cli import (
    PDBeChemManager,
    create_parser,
)
from pdbeccdutils.tests.tst_utilities import test_cut_down_components_cif


class TestCommandLineArgs:
    @staticmethod
    def test_with_empty_args():
        """
        User passes no args, should produce a usage statement and then
        raise SystemExit. Usage statement will appear
        """
        parser = create_parser()
        with pytest.raises(SystemExit):
            parser.parse_args()

    @staticmethod
    def test_input_file_that_cannot_exist_raises_system_exit():
        parser = create_parser()
        with pytest.raises(SystemExit):
            parser.parse_args(["-o foo", "/////impossible_to_open_file", "--debug"])


class TestCutDownComponentsCif:
    """
    run process_components_cif_cli on test file:

    cut_down_components.cif

    that is a cutdown components cif with just the first 5 chemical
    components definitions.

    use pytest fixture to do the run once but then have separate tests
    for the creation of each directory file that is required in the
    ftp area.
    """

    CHEM_COMP_IDS = ["000", "001", "002", "003", "004", "ZPN"]

    @pytest.fixture(scope="class")
    def pipeline_wd(self, tmpdir_factory):
        wd = tmpdir_factory.mktemp("pdbechem_test")

        parser = create_parser()
        args = parser.parse_args(["-o", str(wd), test_cut_down_components_cif])

        m = PDBeChemManager()
        m.run(args.components_cif, args.output_dir)

        return args.output_dir

    @staticmethod
    @pytest.mark.parametrize("chem_comp_id", CHEM_COMP_IDS)
    def test_subdir_tree_created(pipeline_wd, chem_comp_id):
        path = pipeline_wd / chem_comp_id[0] / chem_comp_id
        assert path.is_dir()

    @staticmethod
    @pytest.mark.parametrize("chem_comp_id", CHEM_COMP_IDS)
    def test_all_files_created(pipeline_wd: Path, chem_comp_id):
        path = pipeline_wd / chem_comp_id[0] / chem_comp_id
        files = os.listdir(path)
        assert len(files) == 19

        for f in files:
            file_pointer = path / f
            assert file_pointer.stat().st_size > 0

    @staticmethod
    @pytest.mark.parametrize("id_", ["000", "001", "002", "003", "004"])
    def test_images_with_names_created(pipeline_wd: Path, id_):
        """Test if the depictions with names contain certain atom labels
        as expected.
        """
        path = pipeline_wd / id_[0] / id_ / f"{id_}_100_names.svg"
        xml = ET.parse(path)
        path_elements = xml.findall("svg:path", svg_namespace)

        for e in path_elements:
            if "class" in e.attrib:
                if re.fullmatch(r"atom-\d+", e.attrib["class"]):
                    return  # label is where it should be we can quit

        assert False  # this compound should have had label

    @staticmethod
    @pytest.mark.parametrize(
        "id_,name,alt_name",
        [("001", "H021", "1H02"), ("002", "H121", "1H12"), ("003", "H121", "1H12")],
    )
    def test_correct_atom_naming_in_files(pipeline_wd: Path, id_, name, alt_name):
        """Test if alternate names are used for exported model/ideal
        pdb files.
        """
        alts = [
            pipeline_wd / id_[0] / id_ / f"{id_}_ideal_alt.pdb",
            pipeline_wd / id_[0] / id_ / f"{id_}_model_alt.pdb",
        ]
        regular = [
            pipeline_wd / id_[0] / id_ / f"{id_}_ideal.pdb",
            pipeline_wd / id_[0] / id_ / f"{id_}_model.pdb",
        ]

        for i in alts:
            with open(i) as f:
                str_repr = f.read()
                assert alt_name in str_repr and name not in str_repr

        for i in regular:
            with open(i) as f:
                str_repr = f.read()
                assert name in str_repr and alt_name not in str_repr

    @staticmethod
    @pytest.mark.parametrize("chem_comp_id", CHEM_COMP_IDS)
    def test_cml_files(pipeline_wd: Path, chem_comp_id):
        """Test if the CML file is parsable. Each of the tested compounds
        should contain element with type C.
        """
        path = pipeline_wd / chem_comp_id[0] / chem_comp_id / f"{chem_comp_id}.cml"

        xml_root = ET.parse(str(path)).getroot()
        atoms = xml_root.find("molecule").find("atomArray").findall("atom")

        assert any(a.attrib["elementType"] == "C" for a in atoms)

    @staticmethod
    @pytest.mark.parametrize("chem_comp_id", CHEM_COMP_IDS)
    def test_annotation_file(pipeline_wd: Path, chem_comp_id):
        """Test if the annotation.json is parsable. Each of the tested compounds
        should contain some data.
        """

        path = (
            pipeline_wd
            / chem_comp_id[0]
            / chem_comp_id
            / f"{chem_comp_id}_annotation.json"
        )

        assert path.is_file()

        with open(path, "r") as fp:
            data = json.load(fp)

            assert data
            assert data["atoms"]
            assert data["bonds"]
            assert data["resolution"]
